//----------------------------------------------------------------------------
//
// File:        xgpl.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1998-2000 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include "common.hpp"
#include "logger.hpp"
#include "tms9900.hpp"
#include "tms9918a.hpp"
#include "ti994a.hpp"

#if defined ( LITTLE_ENDIAN )
    #define ENDIAN_ADJUST  1
#else
    #define ENDIAN_ADJUST  0
#endif

cTI994A   *pCPU = NULL;
cTMS9918A *pVDP = NULL;
UCHAR     *pRAM  = NULL;
UCHAR     *pGROM = NULL;
UCHAR     *pVRAM = NULL;

typedef UCHAR *(*gplFunction) ( UCHAR * );

struct gplCode {
    UCHAR       lowOpCode;
    UCHAR       highOpCode;
    gplFunction Function;
};

struct sMoveOpcodeBits {
    UCHAR  NumberIsImmediate    : 1;
    UCHAR  SourceIsGROM         : 1;	// indexed over CPU RAM
    UCHAR  SourceIsVDPorCPU     : 1;
    UCHAR  DestIsVDPRegister    : 1;
    UCHAR  DestIsNotGROM        : 1;
    UCHAR  opcode               : 3;
};

extern "C" UCHAR MemFlags  [ 0x10000 ];
extern sTrapInfo TrapList  [ 16 ];

static void CpuWriteByte ( USHORT address, UCHAR data )
{
    int flags = MemFlags [ address ^ ENDIAN_ADJUST ];
    if ( flags & ( MEMFLG_WRITE | MEMFLG_ROM )) {
        if ( flags & MEMFLG_WRITE ) {
            sTrapInfo *trap = &TrapList [ flags & 0x0F ];
            data = ( UCHAR ) trap->function ( trap->ptr, trap->data, false, address, data );
        }
        if ( flags && MEMFLG_ROM ) return;
    }

    pRAM [ address ^ ENDIAN_ADJUST ] = data;
}

static UCHAR CpuReadByte ( USHORT address )
{
    int flags = MemFlags [ address ^ ENDIAN_ADJUST ];
    UCHAR data = pRAM [ address ^ ENDIAN_ADJUST ];
    if ( flags && MEMFLG_READ ) {
        sTrapInfo *trap = &TrapList [ flags & 0x0F ];
        data = ( UCHAR ) trap->function ( trap->ptr, trap->data, true, address, 0 );
    }
    return data;
}

static void VdpWriteByte ( USHORT, UCHAR data )
{
    pVDP->WriteData ( data );
}

static UCHAR VdpReadByte ( USHORT address )
{
    return pVRAM [ address ];
}

static void GromWriteByte ( USHORT address, UCHAR data )
{
    pGROM [ address ] = data;
}

static UCHAR GromReadByte ( USHORT address )
{
    return pGROM [ address ];
}

void (*WriteTable[])(USHORT, UCHAR) = { CpuWriteByte, VdpWriteByte, GromWriteByte };
UCHAR (*ReadTable[])(USHORT)        = { CpuReadByte,  VdpReadByte,  GromReadByte };

#define MEM_CPU		0x00000000
#define MEM_VDP		0x00010000
#define MEM_GROM	0x00020000

static ULONG gpl_GetSourceDestination ( UCHAR *&opcode )
{
    char temp = *opcode++;

    if ( temp >= 0 ) {
        return MEM_CPU + 0x8300 + temp;
    }

    ULONG base = MEM_CPU;

    USHORT r10 = ( USHORT ) (( temp << 8 ) | *opcode++ );
    USHORT r1 = ( USHORT ) ( r10 & 0x0FFF );

    if ( r1 >= 0x0F00 ) {		// Extended address
        r1 = ( USHORT ) (( r1 << 8 ) | *opcode++ );
    }

    if ( r10 & 0x2000 ) {
        base = MEM_VDP;
        if ( r10 & 0x1000 ) {
            r1 += ( USHORT ) 0x8300;
        }
    } else {
        r1 += ( USHORT ) 0x8300;
    }

    USHORT address = r1;

    if ( r10 & 0x4000 ) {
        address += * ( USHORT * ) ( pRAM + 0x8300 + ( *opcode++ & 0xFFFE ));
    }

    if ( r10 & 0x1000 ) {
        address = * ( USHORT * ) ( pRAM + address );
    }

    return base | address;
}

static ULONG gpl_GetArgument ( UCHAR &mode, UCHAR *&opcode )
{
    int set = mode & 0x80;
    mode <<= 1;

    if ( set ) {
        return gpl_GetSourceDestination ( opcode );
    }

    USHORT address = ( USHORT ) (( opcode[0] << 8 ) | opcode[1] );
    opcode += 2;

    if ( mode & 0x80 ) {
        address += * ( USHORT * ) ( pRAM + 0x8300 + ( *opcode++ & 0xFFFE ));
    }

    return MEM_GROM | address;
}

UCHAR *gpl_RTNC ( UCHAR * )
{
    UCHAR level = pRAM [ 0x8373 ^ ENDIAN_ADJUST ];
    pRAM [ 0x8373 ^ ENDIAN_ADJUST ] -= ( UCHAR ) 2;
    USHORT *stack = ( USHORT * ) ( pRAM + 0x8300 + level );
    return pGROM + *stack;
}

UCHAR *gpl_RTN ( UCHAR *opcode )
{
    pRAM [ 0x837C ^ ENDIAN_ADJUST ] &= ~0x20;
    return gpl_RTNC ( opcode );
}

UCHAR *gpl_B ( UCHAR *opcode )
{
    pRAM [ 0x837C ^ ENDIAN_ADJUST ] &= ~0x20;
    return pGROM + (( opcode [1] << 8 ) | opcode [2] );
}

UCHAR *gpl_ALL ( UCHAR *opcode )
{
    pVDP->WriteAddress ( 0x00 );
    pVDP->WriteAddress ( 0x40 );
    for ( int i = 0; i < 0x0300; i++ ) pVDP->WriteData ( opcode [1] );
    pRAM [ 0x837E ] = 0;
    pRAM [ 0x837F ] = 0;
    return opcode + 2;
}

UCHAR *gpl_BR ( UCHAR *opcode )
{
    int offset = ( opcode - pGROM + 1 ) & 0xE000;
    offset |= (( opcode [0] << 8 ) | opcode [1] ) & 0x1FFF;
    int flags = pRAM [ 0x837C ^ ENDIAN_ADJUST ];
    pRAM [ 0x837C ^ ENDIAN_ADJUST ] &= ~0x20;
    return ( flags & 0x20 ) ? opcode + 2 : pGROM + offset;
}

UCHAR *gpl_BS ( UCHAR *opcode )
{
    int offset = ( opcode - pGROM + 1 ) & 0xE000;
    offset |= (( opcode [0] << 8 ) | opcode [1] ) & 0x1FFF;
    int flags = pRAM [ 0x837C ^ ENDIAN_ADJUST ];
    pRAM [ 0x837C ^ ENDIAN_ADJUST ] &= ~0x20;
    return ( flags & 0x20 ) ? pGROM + offset : opcode + 2;
}

UCHAR *gpl_MOVE ( UCHAR *opcode )
{
    sMoveOpcodeBits *bits = ( sMoveOpcodeBits * ) opcode;
    UCHAR mode = ( UCHAR ) ( *opcode++ << 3 );

    int count;
    if ( bits->NumberIsImmediate ) {
        count = (( USHORT ) opcode[0] << 8 ) | opcode[1];
        opcode += 2;
    } else {
        ULONG address = gpl_GetSourceDestination ( opcode );
        UCHAR (*ReadMemory)(USHORT) = ReadTable [ address >> 16 ];
        count = ( ReadMemory (( USHORT ) address ) << 8 ) | ReadMemory (( USHORT ( address + 1 )));
    }

    ULONG dest = gpl_GetArgument ( mode, opcode );
    mode <<= 1;
    ULONG src = gpl_GetArgument ( mode, opcode );

    void (*WriteMemory) (USHORT, UCHAR) = WriteTable [ dest >> 16 ];
    UCHAR (*ReadMemory) (USHORT) = ReadTable [ src >> 16 ];

    if ( bits->DestIsVDPRegister ) {
        USHORT srcAddress = ( USHORT ) src;
        int reg = dest - 0x8300;
        for ( int i = 0; i < count; i++ ) {
            pCPU->GetVDP ()->WriteRegister ( reg + i, ReadMemory ( srcAddress++ ));
        }
    } else {
        USHORT srcAddress = ( USHORT ) src;
        USHORT destAddress = ( USHORT ) dest;
        for ( int i = 0; i < count; i++ ) {
            WriteMemory ( destAddress++, ReadMemory ( srcAddress++ ));
        }
        if ( src >> 16 ) {
            // set grom/vdp read address
        }
        if ( dest >> 16 ) {
            // set grom/vdp write address
        }
    }

    return opcode;
}

static gplCode GPL [] = {
  { /* RTN   */ 0x00, 0x00, gpl_RTN },
  { /* RTNC  */ 0x01, 0x01, gpl_RTNC },
  { /* RAND  */ 0x02, 0x02, NULL },
  { /* SCAN  */ 0x03, 0x03, NULL },
  { /* BACK  */ 0x04, 0x04, NULL },
  { /* B     */ 0x05, 0x05, gpl_B },
  { /* CALL  */ 0x06, 0x06, NULL },
  { /* ALL   */ 0x07, 0x07, gpl_ALL },
  { /* FMT   */ 0x08, 0x08, NULL },
  { /* H     */ 0x09, 0x09, NULL },
  { /* GT    */ 0x0A, 0x0A, NULL },
  { /* EXIT  */ 0x0B, 0x0B, NULL },
  { /* CARRY */ 0x0C, 0x0C, NULL },
  { /* OVF   */ 0x0D, 0x0D, NULL },
  { /* PARSE */ 0x0E, 0x0E, NULL },
  { /* XML   */ 0x0F, 0x0F, NULL },
  { /* CONT  */ 0x10, 0x10, NULL },
  { /* EXEC  */ 0x11, 0x11, NULL },
  { /* RTNB  */ 0x12, 0x12, NULL },
  { /* RTGR  */ 0x13, 0x13, NULL },
  { /* MOVE  */ 0x20, 0x3F, NULL },
//  { /* MOVE  */ 0x20, 0x3F, gpl_MOVE },
  { /* BR    */ 0x40, 0x5F, gpl_BR },
  { /* BS    */ 0x60, 0x7F, gpl_BS },
  { /* ABS   */ 0x80, 0x81, NULL },
  { /* NEG   */ 0x82, 0x83, NULL },
  { /* INV   */ 0x84, 0x85, NULL },
  { /* CLR   */ 0x86, 0x87, NULL },
  { /* FETCH */ 0x88, 0x89, NULL },
  { /* CASE  */ 0x8A, 0x8B, NULL },
  { /* PUSH  */ 0x8C, 0x8D, NULL },
  { /* CZ    */ 0x8E, 0x8F, NULL },
  { /* INC   */ 0x90, 0x91, NULL },
  { /* DEC   */ 0x92, 0x93, NULL },
  { /* INCT  */ 0x94, 0x95, NULL },
  { /* DECT  */ 0x96, 0x97, NULL },
  { /* ADD   */ 0xA0, 0xA3, NULL },
  { /* SUB   */ 0xA4, 0xA7, NULL },
  { /* MUL   */ 0xA8, 0xAB, NULL },
  { /* DIV   */ 0xAC, 0xAF, NULL },
  { /* AND   */ 0xB0, 0xB3, NULL },
  { /* OR    */ 0xB4, 0xB7, NULL },
  { /* XOR   */ 0xB8, 0xBB, NULL },
  { /* ST    */ 0xBC, 0xBF, NULL },
  { /* EX    */ 0xC0, 0xC3, NULL },
  { /* CH    */ 0xC4, 0xC7, NULL },
  { /* CHE   */ 0xC8, 0xCB, NULL },
  { /* CGT   */ 0xCC, 0xCF, NULL },
  { /* CGE   */ 0xD0, 0xD3, NULL },
  { /* CEQ   */ 0xD4, 0xD7, NULL },
  { /* CLOG  */ 0xD8, 0xDB, NULL },
  { /* SRA   */ 0xDC, 0xDF, NULL },
  { /* SLL   */ 0xE0, 0xE3, NULL },
  { /* SRL   */ 0xE4, 0xE7, NULL },
  { /* SRC   */ 0xE8, 0xEB, NULL },
  { /* COINC */ 0xED, 0xED, NULL },
  { /* I/O   */ 0xF4, 0xF7, NULL },
  { /* SWGR  */ 0xF8, 0xFB, NULL },
  { /* UNKN  */ 0x00, 0xFF, NULL }
};

void InitGPL ( cTI994A *cpu )
{
    pCPU = cpu;
    pVDP = pCPU->GetVDP ();
    pRAM = pCPU->GetCpuMemory ();
    pGROM = pCPU->GetGromMemory ();
    pVRAM = pCPU->GetVideoMemory ();
}

extern "C" bool ExecuteGPL ()
{
    gplCode *gpl = GPL;

    UCHAR *opcode = pGROM + pCPU->GetGromAddress ();

    for ( unsigned i = 0; i < SIZE ( GPL ); i++, gpl++ ) {
        if (( *opcode >= gpl->lowOpCode ) && ( *opcode <= gpl->highOpCode )) break;
    }

    if ( gpl->Function == NULL ) {
        return false;
    }

    opcode = gpl->Function ( opcode );

    pCPU->SetGromAddress (( USHORT ) ( opcode - pGROM ));

    return true;
}
